import os
import torch
import random
from dataclasses import dataclass
import diffusers
import torch.utils
import yaml
import argparse

import unet.utils as utils
from unet.unet_controller import UNetController
from PIL import Image

diffusers.utils.logging.set_verbosity_error()

@dataclass
class TestConfig:
    model_path: str = None
    project_base_path: str = '.'
    base_save_dir: str = None
    torch_dtype: torch.dtype = torch.float16
    variant: str = "fp16"
    device: str = None
    seed: int = 42
    yaml_path: str = './resource/ConsiStory+.yaml'

def parse_args():
    parser = argparse.ArgumentParser(description="Generate images with UNetController")
    
    parser.add_argument('--model_path', type=str, default=TestConfig.model_path, help='Path to the model')
    parser.add_argument('--mode', type=str, nargs='+', help='List of modes to run')
    parser.add_argument('--project_base_path', type=str, default=TestConfig.project_base_path, help='Base path of the project')
    parser.add_argument('--base_save_dir', type=str, default=TestConfig.base_save_dir, help='Base directory to save results')
    parser.add_argument('--torch_dtype', type=lambda x: getattr(torch, x), default=TestConfig.torch_dtype, help='Torch data type')
    parser.add_argument('--variant', type=str, default=TestConfig.variant, help='Variant type')
    parser.add_argument('--device', type=str, default=TestConfig.device, help='Device to run on')
    parser.add_argument('--seed', type=int, default=TestConfig.seed, help='Random seed')
    parser.add_argument('--yaml_path', type=str, default=TestConfig.yaml_path, help='yaml_path')
    parser.add_argument('--window_length', type=int, default=2)
    parser.add_argument('--single', action='store_true', help='single_folder')
    parser.add_argument('--gen_num', type=int, default=1,)

    return parser.parse_args()

if __name__ == "__main__":
    args = parse_args()

    # Load model
    pipe, model_name = utils.load_pipe_from_path(args.model_path, args.device, args.torch_dtype, args.variant)

    if not os.path.exists(args.base_save_dir):
        os.mkdir(args.base_save_dir)

    with open(args.yaml_path, 'r') as file:
        data = yaml.safe_load(file)

    for i in range(args.gen_num):
        args.seed = random.randint(1, 10000)
        # args.seed = 5830
        print(args.seed)
        for mode in args.mode:

            save_dir = os.path.join(args.base_save_dir, mode)
            os.makedirs(save_dir, exist_ok=True)

            unet_controller = UNetController()
            unet_controller.set_device(args.device)
            unet_controller.set_tokenizer(pipe.tokenizer)
            unet_controller.save_intermediate_result = False
            unet_controller.is_freeu_enabled = True
            unet_controller.use_sdsa = True
            unet_controller.sdsa_position = ['up0']
            unet_controller.sdca_position = ['up0']

            unet_controller.set_variable('use_same_init_noise', True)
            unet_controller.set_variable('single', False)
            unet_controller.set_variable('sdca2', False)

            # Set prompt embedding modes based on the current mode
            if mode == "npr":
                unet_controller.prompt_embeds_mode = 'weighting enhance and weaken'
                unet_controller.use_sdca = False
                unet_controller.use_sdsa = False
            elif mode == "npr+":
                unet_controller.prompt_embeds_mode = 'weighting enhance'
                unet_controller.use_sdca = False
                unet_controller.use_sdsa = False
            elif mode == "npr-":
                unet_controller.prompt_embeds_mode = 'weighting weaken'
                unet_controller.use_sdca = False
                unet_controller.use_sdsa = False    
            elif mode == "svr+":
                unet_controller.prompt_embeds_mode = 'only enhance'
                unet_controller.use_sdca = False
            elif mode == "svr-":
                unet_controller.prompt_embeds_mode = 'only suppress'
                unet_controller.use_sdca = False
            elif mode == "ipca":
                unet_controller.prompt_embeds_mode = 'original'
                unet_controller.use_sdca = True
            elif mode == "svr+ & svr-":
                unet_controller.prompt_embeds_mode = 'enhance and suppress'
                unet_controller.use_sdca = False
            elif mode == "svr- & svr+":
                unet_controller.prompt_embeds_mode = 'suppress and enhance'
                unet_controller.use_sdca = False
            elif mode == "svr+ & ipca":
                unet_controller.prompt_embeds_mode = 'only enhance'
                unet_controller.use_sdca = True
            elif mode == "svr- & ipca":
                unet_controller.prompt_embeds_mode = 'only suppress'
                unet_controller.use_sdca = True
            elif mode == "svr+ & svr- & ipca" or mode == "ours":
                unet_controller.prompt_embeds_mode = 'enhance and suppress'
                unet_controller.use_sdca = True
            elif mode == "original":
                unet_controller.use_sdsa = False
                unet_controller.use_sdca = False
                unet_controller.prompt_embeds_mode = 'original'
                unet_controller.set_variable('use_same_init_noise', False)
                unet_controller.set_variable('original', True)

            else:
                print(f"Unknown mode: {mode}")
                continue

            # Generate images
            if not args.single:
                for subject_domain, subject_domain_instances in data.items():
                    for index, instance in enumerate(subject_domain_instances):
                        id_prompt = f'{instance["style"]} {instance["subject"]}'
                        movement_prompt_list = instance["settings"]
                        keyword_list = [f'{instance["concept_token"]}']
                        args.seed = random.randint(1, 10000)

                        save_path = os.path.join(save_dir, f"{keyword_list[0]}_{args.seed}")
                        os.makedirs(save_path, exist_ok=True)

                        unet_controller.set_variable('seed', args.seed)
                        unet_controller.set_keyword_list(keyword_list)
                        images, combine_image = utils.movement_gen_story_slide_windows(id_prompt, movement_prompt_list, pipe, args.window_length, args.seed, unet_controller)


                        for movement, image in zip(movement_prompt_list, images):
                            image.save(os.path.join(save_path, f'{id_prompt} {movement}.jpg'))
                        combine_image.save(os.path.join(args.base_save_dir, f"{keyword_list[0]}_{args.seed}.jpg"))
            else:
                # id_prompt = "A vintage-style poster of A dog"
                # movement_prompt_list = [
                #     "playing a guitar at a country concert",
                #     "sitting by a campfire under a starry sky",
                #     "riding a skateboard through a bustling city",
                #     "posing in front of a historical landmark",
                #     "wearing an astronaut suit on the moon"
                # ]
                # keyword_list = ['dog']

                id_prompt = "A vintage-style poster of A dog"
                movement_prompt_list = [
                    "dancing to music at a vibrant street festival",
                    "chasing a frisbee in a colorful park",
                    "wearing sunglasses while relaxing on a beach chair",
                    "posing for a photoshoot in a modern art gallery",
                    "jumping through a hoop at a circus performance",
                    "playing with a group of children at a playground",
                    "exploring a retro diner while wearing a bowtie"
                ]
                keyword_list = ['dog']

                # id_prompt = "A mystical illustration of A wise wizard with a long, flowing beard"
                # movement_prompt_list = [
                #     "in a tower filled with ancient tomes and artifacts",
                #     "casting a spell by the light of a full moon",
                #     "standing before a magical portal in the forest",
                #     "summoning a storm over a mountain peak",
                #     "writing runes in a dusty spellbook",
                #     "mixing potions in a dimly lit chamber",
                #     "consulting a crystal ball"
                # ]
                # keyword_list = ['wizard']

                # id_prompt: str = "A pixar style illustration of a dragon"
                # keyword_list = ["dragon"]
                # movement_prompt_list = [
                #         "soaring gracefully through a rainbow sky",
                #         "nestled among blooming cherry blossoms",
                #         "playfully splashing in a sparkling lake",
                # ]

                # id_prompt: str = "A hyper-realistic digital painting of a 16 years old girl."
                # keyword_list = ["girl"]
                # movement_prompt_list = [
                #     "in a flower garden",
                #     "building a sandcastle",
                #     "in a city park with autumn leaves",
                # ]

                # id_prompt: str = "A whimsical painting of A delicate fairy"
                # keyword_list = ["fairy"]
                # movement_prompt_list = [
                #     "hovering over a moonlit pond",
                #     "dancing on the petals of a giant flower",
                #     "hiding in the hollow of an ancient tree",
                #     "spreading fairy dust over a sleeping village",
                #     "sitting on a mushroom in a magical forest",
                #     "playing with fireflies at dusk",
                # ]

                # id_prompt: str = "A hyper-realistic digital painting of An elderly gentleman"
                # keyword_list = ["gentleman"]
                # movement_prompt_list = [
                #     "wearing a smoking jacket",
                #     "at a vintage car show",
                #     "wearing a vineyard owner's attire",
                #     "on a golf course",
                #     "at a classical music concert",
                #     "painting a landscape",
                # ]


                # id_prompt: str = "A vintage-style poster of A ceramic vase with an intricate floral pattern and a glossy, sky-blue glaze"
                # keyword_list = ["vase"]
                # movement_prompt_list = [
                #     "holding a rare bouquet of flowers",
                #     "displaying exotic orchids",
                #     "adding charm to a homely setting",
                #     "complementing a corporate decor",
                #     "containing delicate cherry blossoms",
                #     "gracing an elegant dinner table",
                #     "holding a vibrant arrangement of sunflowers",
                #     "filled with a fresh bouquet of lavender and wild daisies"
                # ]

                # id_prompt: str = "A photo of A happy hedgehog with its cheese"
                # keyword_list = ["hedgehog","cheese"]
                # movement_prompt_list = [
                #     "in a garden",
                #     "in an autumn forest",
                #     "next to a tiny cheese wheel",
                #     "sitting on a mushroom",
                #     "under a picnic blanket",
                #     "beside a sparkling stream",
                #     "amid blooming spring flowers",
                #     "peeking from a cozy burrow"
                # ]

                # id_prompt: str = "A heartwarming illustration of A friendly troll"
                # keyword_list = ["troll"]
                # movement_prompt_list = [
                #     "under a stone bridge covered in ivy",
                #     "guarding a treasure chest in a dark cave",
                #     "helping travelers across a river",
                #     "sitting by a campfire in a foggy forest",
                #     "building a shelter from fallen logs",
                #     "fishing in a quiet stream at dusk",
                #     "carving runes into a rock",
                #     "resting under a large oak tree"
                # ]

                # id_prompt: str = "A contemporary minimalist sketch of A tall cactus with vibrant blooms"
                # keyword_list = ["cactus"]
                # movement_prompt_list = [
                #     "in a sun-drenched desert landscape",
                #     "against a backdrop of red rock formations",
                #     "with a blazing sunset in the background",
                #     "standing alone in an arid, sandy terrain",
                #     "with a clear blue sky overhead",
                #     "amidst a field of dry grasses",
                #     "casting a long shadow on the desert floor"
                # ]

                # id_prompt: str = "A quaint illustration of A hobbit"
                # keyword_list = ["hobbit"]
                # movement_prompt_list = [
                #     "in a cozy, round door cottage",
                #     "sitting by a fireplace in a quaint home",
                #     "working in a garden of vibrant vegetables",
                #     "enjoying a feast under a starlit sky",
                #     "reading a book in a sunlit meadow",
                #     "walking through a peaceful village",
                #     "celebrating with friends in a rustic tavern",
                #     "exploring a hidden valley"
                # ]

                # id_prompt: str = "A hyper-realistic digital painting of a young ginger boy with his ball"
                # keyword_list = ["boy","ball"]
                # movement_prompt_list = [
                #     "leaves scattering in a gentle breeze",
                #     "standing in a quiet meadow",
                #     "set against a vibrant sunset",
                #     "near a bustling playground",
                #     "by a serene reflecting lake",
                #     "in a busy street of people",
                #     "by a colorful graffiti wall",
                #     "amidst a field of blooming wildflowers",
                #     "on a quiet forest trail after rain"
                # ]

                # id_prompt: str = "A cinematic portrait of a man and a woman standing together"
                # keyword_list = ["man", "woman"]
                # movement_prompt_list = [
                #     "under a sky full of stars",
                #     "on a bustling city street at night",
                #     "in a dimly lit jazz club",
                #     "walking along a sandy beach at sunset",
                #     "in a cozy coffee shop with large windows",
                #     "in a vibrant art gallery surrounded by paintings",
                #     "under an umbrella during a soft rain",
                #     "on a quiet park bench amidst falling leaves",
                #     "standing on a rooftop overlooking the city skyline"
                # ]

                # id_prompt: str = "A cinematic portrait of a man, a woman, and a child"
                # keyword_list = ["man", "woman", "child"]
                # movement_prompt_list = [
                #     "walking in a quiet park",
                #     "under a starlit sky",
                #     "by a rustic cabin",
                #     "on a forest trail",
                #     "by a peaceful lake",
                #     "at a vibrant market",
                #     "in a snowy street",
                #     "by a carousel",
                #     "on a picnic blanket"
                # ]

                # id_prompt: str = "A cinematic portrait of a father, a mother, and a child"
                # keyword_list = ["father", "mother", "child"]
                # movement_prompt_list = [
                #     "walking in a quiet park",
                #     "under a starlit sky",
                #     "by a rustic cabin",
                #     "on a forest trail",
                #     "by a peaceful lake",
                #     "at a vibrant market",
                #     "in a snowy street",
                #     "by a carousel",
                #     "on a picnic blanket"
                # ]

                # id_prompt: str = "A cinematic portrait of a tall man, a playful dog, and a curious cat"
                # keyword_list = ["man", "dog", "cat"]
                # movement_prompt_list = [
                #     "walking through a quiet park",
                #     "resting under a large oak tree",
                #     "by a cozy fireplace",
                #     "sitting on a sunny porch",
                #     "by a calm lake at dawn",
                #     "in a bustling town square",
                #     "on a winding forest trail",
                #     "at a peaceful riverside",
                #     "on a soft blanket in the garden"
                # ]



                save_path = os.path.join(save_dir, f"{keyword_list[0]}_{args.seed}")
                os.makedirs(save_path, exist_ok=True)

                unet_controller.set_variable('seed', args.seed)
                unet_controller.set_keyword_list(keyword_list)
                images, combine_image = utils.movement_gen_story_slide_windows(id_prompt, movement_prompt_list, pipe, args.window_length, args.seed, unet_controller)


                for movement, image in zip(movement_prompt_list, images):
                    image.save(os.path.join(save_path, f'{id_prompt} {movement}.jpg'))
                combine_image.save(os.path.join(save_path, f"{keyword_list[0]}_{args.seed}.jpg"))
