from typing import Optional
import torch
import os 

import unet.utils as utils

class UNetController():
    def __init__(self):
        self._variables = {}
        self.torch_dtype = torch.float16
        self.is_freeu_enabled = False
        self.freeu_parm = {'s1':0.6, 's2':0.4, 'b1':1.1, 'b2':1.2}

        self.device = "cuda"
        self.current_unet_position = 'down' # down, mid or up
        
        self.width = 1024
        self.height = 1024
        self.prompts = None
        self.tokenizer = None
        self.ref_prompts = None
        self.keyword_list = None
        self.current_latents = None
        self.result_save_dir = None 
        self.negative_prompt = None
        self.current_time_step = None
        self.total_avg_attn_map = None
        self.do_classifier_free_guidance = None

        self.save_intermediate_result = False
        self.save_self_attn_map = True
        self.save_self_attn_mask = True
        self.save_cross_attn_map = True
        self.save_cross_attn_mask = True

        # embedding editting
        self.prompt_embeds_mode = 'enhance and suppress' 
        self.use_same_latents = True
        self.remove_pool_embeds = False
        self.prompt_embeds_start_step = 0
        self.optimize_embeds = False
        self.learning_rate = 1e-5
        self.optimization_steps = 1

        # sdsa
        self.use_sdsa = False
        self.sdsa_position = ['up0','up1','up2']
        self.sdsa_start_step = 0
        self.sdsa_dropout = 0.5

        # sdca
        self.use_sdca = False
        self.sdca_position = ['up0','up1','up2']
        self.sdca_start_step = 0
        self.sdca_dropout = 0.5

        # weak story attn-map
        self.weak_self_attn_story_prompt = False
        self.weak_cross_attn_story_prompt = False
        self.self_attn_story_prompt_dropout = 0.5
        self.cross_attn_story_prompt_dropout = 0.5

        # cross-attn-mask
        self.use_cross_attn_mask = True
        self.threshold_mode = 'manual'  # 0:ostu 1:manual
        self.attn_mask_mode = 'horizontal' #
        self.cross_attn_mask_start_step = 0
        self.cross_attn_threshold = 0.4
        self.cross_attn_map_store_size = 32 ** 2 # only store the size in 32 ** 2


        print("------------init_unet_controller--------------")
        # self.print_attributes()

    def set_variable(self, name, value):
        self._variables[name] = value

    def get_variable(self, name):
        return self._variables.get(name)
    
    def variable_exists(self, name):
        return name in self._variables

    def reinit_for_next_prompts(self):
        self.total_avg_attn_map = None
        self.save_self_attn_map = True
        self.save_self_attn_mask = True
        self.save_cross_attn_map = True
        self.save_cross_attn_mask = True
    
    def update_state(self, current_time_step):
        self.current_time_step = current_time_step

    def add_current_cross_attn_map(self, cross_attn_map):
        cross_attn_map = torch.sum(cross_attn_map, dim=1) * self.current_time_step / 50.0
        if self.total_avg_attn_map is None:
            self.total_avg_attn_map = cross_attn_map
        else:
            self.total_avg_attn_map = self.total_avg_attn_map + cross_attn_map

    def set_tokenizer(self, tokenizer):
        self.tokenizer = tokenizer

    def set_ref_prompts(self, ref_prompts):
        self.ref_prompts = ref_prompts

    def set_result_save_dir(self, result_save_dir, mkdir=False):
        self.result_save_dir = result_save_dir
        if mkdir is True and os.path.exists(result_save_dir) is False:
            os.mkdir(result_save_dir)

    def set_keyword_list(self, keyword_list):
        self.keyword_list = keyword_list

    def set_device(self, device):
        self.device = device

    def set_prompts(self, prompts):
        self.prompts = prompts

    def set_negative_prompt(self, negative_prompt):
        self.negative_prompt = negative_prompt

    def set_current_time_step(self, t):
        self.current_time_step = t

    def set_current_latents(self, latents):
        self.current_latents = latents

    def set_do_classifier_free_guidance(self, do_classifier_free_guidance):
        self.do_classifier_free_guidance = do_classifier_free_guidance

    def set_current_unet_position(self, current_unet_position):
            self.current_unet_position = current_unet_position
    
    def eval(self, extra_output:Optional[str] = None):
        # print(f'current_t:{self.current_time_step}')
        if extra_output != None:
            print(extra_output)

    def print_attributes(self):
        """
        Prints all attributes and their values of the object.
        """
        for attr, value in vars(self).items():
            print(f"{attr}: {value}")

    def get_cross_attn_map_by_keyword(self, keyword, prompt_index, get_positive_branch=True, reshape=True, sum=False):
        if prompt_index > (len(self.prompts) - 1):
            exit("propmt_index out of range")
        
        keyword_indices = []
        prompt = self.prompts[prompt_index]
        prompt_tokens = utils.prompt2tokens(self.tokenizer,prompt)
        
        keyword_tokens = utils.prompt2tokens(self.tokenizer,keyword)
        keyword_tokens = [word for word in keyword_tokens if word != '<|endoftext|>' and word != '<|startoftext|>']
        index_of_words = utils.find_sublist_index(prompt_tokens,keyword_tokens)
        keyword_indices.extend([num for num in range(index_of_words, index_of_words+len(keyword_tokens))])

        cross_attn_map_neg, cross_attn_map_pos = torch.split(self.total_avg_attn_map, self.total_avg_attn_map.size(0) // 2, dim=0)

        if get_positive_branch:
            cross_attn_map = cross_attn_map_pos[prompt_index]
        else:
            cross_attn_map = cross_attn_map_neg[prompt_index]
        cross_attn_map_tran = torch.transpose(cross_attn_map,0,1)
        keyword_attn_map = cross_attn_map_tran[keyword_indices]
        if sum:
            keyword_attn_map = torch.sum(keyword_attn_map, dim=0).unsqueeze(0)

        side_length = int(keyword_attn_map.size(1) ** 0.5)
        if reshape:
            keyword_attn_map = keyword_attn_map.view(1,side_length,side_length)

        return keyword_attn_map 

        