import torch
from typing import Optional
from PIL import Image
from diffusers import AutoencoderKL, EulerDiscreteScheduler, EDMDPMSolverMultistepScheduler
from transformers import (
    CLIPTextModel,
    CLIPTextModelWithProjection,
    CLIPTokenizer,
)
from scipy.spatial.distance import cdist
import torchvision.transforms as transforms
import numpy as np
import cv2
import os
import random
import unet.pipeline_stable_diffusion_xl as pipeline_stable_diffusion_xl
from unet.unet import UNet2DConditionModel
from unet.unet_controller import UNetController
from torch.fft import fftn, fftshift, ifftn, ifftshift
from typing import List, Optional, Tuple, Union



def sdsa(q, k, v, scale, unet_controller: Optional[UNetController] = None): # eg. q,k,v: [4,20,1024,64]
    q_neg, q_pos = torch.split(q, q.size(0) // 2, dim=0)
    k_neg, k_pos = torch.split(k, k.size(0) // 2, dim=0)
    v_neg, v_pos = torch.split(v, v.size(0) // 2, dim=0)

    # 1. negative_attn

    scores_neg = torch.matmul(q_neg, k_neg.transpose(-2, -1)) * scale
    attn_weights_neg = torch.softmax(scores_neg, dim=-1)
    attn_output_neg = torch.matmul(attn_weights_neg, v_neg)

    # 2. positive_attn (we do sdsa only on positive branch)

    # 2.1 sdsa 
    k_plus = torch.cat(tuple(k_pos.transpose(-2, -1)), dim=2).unsqueeze(0).repeat(k_pos.size(0),1,1,1) # 𝐾+ = [𝐾1 ⊕ 𝐾2 ⊕ . . . ⊕ 𝐾𝑁 ]
    v_plus = torch.cat(tuple(v_pos), dim=1).unsqueeze(0).repeat(v_pos.size(0),1,1,1) # 𝑉+ = [𝑉1 ⊕ 𝑉2 ⊕ . . . ⊕ 𝑉𝑁 ]

    # visuialize: qkv_pos_save
    if unet_controller is not None and unet_controller.save_intermediate_result and (unet_controller.current_time_step + 1) % 5 == 0:
        qkv_save_dir = os.path.join(unet_controller.result_save_dir,f"self_qkv")
        if os.path.exists(qkv_save_dir) is False:
            os.mkdir(qkv_save_dir)

        for prompt_index in range(len(unet_controller.prompts)):
            k_pos_save = torch.sum(k_pos[prompt_index],dim=0)
            k_pos_save = torch.sum(k_pos_save,dim=1)
            k_pos_save = k_pos_save.view(int(k_pos_save.size(0)**0.5),int(k_pos_save.size(0)**0.5))
            save_tensor_as_image(k_pos_save,f'{qkv_save_dir}/k_prompt{prompt_index}_{unet_controller.current_time_step}.jpg')

            q_pos_save = torch.sum(q_pos[prompt_index],dim=0)
            q_pos_save = torch.sum(q_pos_save,dim=1)
            q_pos_save = q_pos_save.view(int(q_pos_save.size(0)**0.5),int(q_pos_save.size(0)**0.5))
            save_tensor_as_image(q_pos_save,f'{qkv_save_dir}/q_prompt{prompt_index}_{unet_controller.current_time_step}.jpg')
    

    # 2.2 apply mask
    if unet_controller is not None:
        scores_pos = torch.matmul(q_pos, k_plus) * scale

        # visuialize: attn_map_save
        if unet_controller.save_self_attn_map and unet_controller.save_intermediate_result and unet_controller.current_time_step == 49:
            map_save_dir = os.path.join(unet_controller.result_save_dir,f"self_attn_map")
            if os.path.exists(map_save_dir) is False:
                os.mkdir(map_save_dir)
            side_length = scores_pos.size(2)
            attn_map_save = torch.sum(scores_pos, dim=1)
            for prompt_index in range(len(unet_controller.prompts)):
                attn_map_prompt = attn_map_save[prompt_index,:,prompt_index*side_length:(prompt_index+1)*side_length]
                attn_map_prompt = self_attn_map_reshape(attn_map_prompt)
                save_tensor_as_image(attn_map_prompt,f'{map_save_dir}/{prompt_index}_{unet_controller.current_time_step}.jpg')
            unet_controller.save_self_attn_map = False
        


        # 2.2.1 apply dropout mask
        dropout_mask = gen_dropout_mask(scores_pos.shape, unet_controller, unet_controller.sdsa_dropout) # eg: [a,1024,2048]   

        # 2.2.2 weak story_self_attn_map
        if unet_controller.ref_prompts is not None and unet_controller.weak_self_attn_story_prompt:
            weak_story_prompt_self_attn_map(0,dropout_mask,unet_controller)

        # 2.2.3 apply attn_mask
        if (unet_controller.use_cross_attn_mask and unet_controller.keyword_list is not None and unet_controller.current_time_step >= unet_controller.cross_attn_mask_start_step
            and unet_controller.total_avg_attn_map is not None):
            if unet_controller.attn_mask_mode == 'original':
                cross_attn_mask = gen_cross_attn_mask_0(unet_controller, k.size(2))  
            elif unet_controller.attn_mask_mode == 'horizontal':
                cross_attn_mask = gen_cross_attn_mask_1(unet_controller, k.size(2))
            elif unet_controller.attn_mask_mode == 'horizontal and vertical':
                cross_attn_mask = gen_cross_attn_mask_2(unet_controller, k.size(2))
            elif unet_controller.attn_mask_mode == 'reshape':
                cross_attn_mask = gen_cross_attn_mask_3(unet_controller, k.size(2))
            else:
                exit("attn_mask_mode error")

            mask = dropout_mask * cross_attn_mask
        else:
            mask = dropout_mask

        # visuialize: attn_mask_save
        if unet_controller.save_self_attn_mask and unet_controller.save_intermediate_result and unet_controller.current_time_step == 49:
            map_save_dir = os.path.join(unet_controller.result_save_dir,f"self_attn_mask")
            if os.path.exists(map_save_dir) is False:
                os.mkdir(map_save_dir)
            for prompt_index in range(len(unet_controller.prompts)):
                save_tensor_as_image(mask[prompt_index],f'{map_save_dir}/{prompt_index}_{unet_controller.current_time_step}.jpg')
            unet_controller.save_self_attn_mask = False # only save once

        mask = mask.unsqueeze(1).repeat(1,scores_pos.size(1),1,1)

        attn_weights_pos = torch.softmax(scores_pos + torch.log(mask), dim=-1)
    else:
        scores_pos = torch.matmul(q_pos, k_plus) * scale
        attn_weights_pos = torch.softmax(scores_pos, dim=-1)


    attn_output_pos = torch.matmul(attn_weights_pos, v_plus)

    # 3. combine
    attn_output = torch.cat((attn_output_neg, attn_output_pos), dim=0)

    return attn_output


def sdca(q, k, v, scale, unet_controller: Optional[UNetController] = None): # eg. q: [4,20,1024,64] k,v: [4,20,77,64] 
    q_neg, q_pos = torch.split(q, q.size(0) // 2, dim=0)
    k_neg, k_pos = torch.split(k, k.size(0) // 2, dim=0)
    v_neg, v_pos = torch.split(v, v.size(0) // 2, dim=0)

    # 1. negative_attn

    scores_neg = torch.matmul(q_neg, k_neg.transpose(-2, -1)) * scale
    attn_weights_neg = torch.softmax(scores_neg, dim=-1)
    attn_output_neg = torch.matmul(attn_weights_neg, v_neg)

    # 2. positive_attn (we do sdsa only on positive branch)

    # 2.1 sdca 
    k_plus = torch.cat(tuple(k_pos.transpose(-2, -1)), dim=2).unsqueeze(0).repeat(k_pos.size(0),1,1,1) # 𝐾+ = [𝐾1 ⊕ 𝐾2 ⊕ . . . ⊕ 𝐾𝑁 ]
    v_plus = torch.cat(tuple(v_pos), dim=1).unsqueeze(0).repeat(v_pos.size(0),1,1,1) # 𝑉+ = [𝑉1 ⊕ 𝑉2 ⊕ . . . ⊕ 𝑉𝑁 ]

    # visuialize: qkv_pos_save
    if unet_controller is not None and unet_controller.save_intermediate_result and (unet_controller.current_time_step + 1) % 5 == 0:
        qkv_save_dir = os.path.join(unet_controller.result_save_dir,f"cross_qkv")
        if os.path.exists(qkv_save_dir) is False:
            os.mkdir(qkv_save_dir)

        for prompt_index in range(len(unet_controller.prompts)):

            q_pos_save = torch.sum(q_pos[prompt_index],dim=0)
            q_pos_save = torch.sum(q_pos_save,dim=1)
            q_pos_save = q_pos_save.view(int(q_pos_save.size(0)**0.5),int(q_pos_save.size(0)**0.5))
            save_tensor_as_image(q_pos_save,f'{qkv_save_dir}/q_prompt{prompt_index}_{unet_controller.current_time_step}.jpg')
    

    # 2.2 apply mask
    if unet_controller is not None:
        scores_pos = torch.matmul(q_pos, k_plus) * scale

        # visuialize: attn_map_save
        if unet_controller.save_cross_attn_map and unet_controller.save_intermediate_result and unet_controller.current_time_step == 49:
            map_save_dir = os.path.join(unet_controller.result_save_dir,f"cross_attn_map")
            if os.path.exists(map_save_dir) is False:
                os.mkdir(map_save_dir)
            side_length = int(scores_pos.size(3)/len(unet_controller.prompts))
            attn_map_save = torch.sum(scores_pos, dim=1)
            for prompt_index in range(len(unet_controller.prompts)):
                attn_map_prompt = attn_map_save[prompt_index,:,prompt_index*side_length:(prompt_index+1)*side_length]
                save_tensor_as_image(attn_map_prompt,f'{map_save_dir}/{prompt_index}_{unet_controller.current_time_step}.jpg')
            unet_controller.save_cross_attn_map = False
        

        # 2.2.1 apply dropout mask
        dropout_mask = gen_dropout_mask(scores_pos.shape, unet_controller, unet_controller.sdca_dropout) # eg: [a,1024,154]   

        # 2.2.2 weak story_self_attn_map
        if unet_controller.ref_prompts is not None and unet_controller.weak_cross_attn_story_prompt:
            weak_story_prompt_cross_attn_map(0,dropout_mask,unet_controller)

        # 2.2.3 apply embeds mask (To Do)

        mask = dropout_mask

        # visuialize: attn_mask_save
        if unet_controller.save_cross_attn_mask and unet_controller.save_intermediate_result and unet_controller.current_time_step == 49:
            map_save_dir = os.path.join(unet_controller.result_save_dir,f"cross_attn_mask")
            if os.path.exists(map_save_dir) is False:
                os.mkdir(map_save_dir)
            for prompt_index in range(len(unet_controller.prompts)):
                save_tensor_as_image(mask[prompt_index],f'{map_save_dir}/{prompt_index}_{unet_controller.current_time_step}.jpg')
            unet_controller.save_cross_attn_mask = False # only save once

        mask = mask.unsqueeze(1).repeat(1,scores_pos.size(1),1,1)
        attn_weights_pos = torch.softmax(scores_pos + torch.log(mask), dim=-1)

    else:
        scores_pos = torch.matmul(q_pos, k_plus) * scale
        attn_weights_pos = torch.softmax(scores_pos, dim=-1)


    attn_output_pos = torch.matmul(attn_weights_pos, v_plus)
    # 3. combine
    attn_output = torch.cat((attn_output_neg, attn_output_pos), dim=0)

    return attn_output

def sdca2(q, k, v, scale, unet_controller: Optional[UNetController] = None): # eg. q: [4,20,1024,64] k,v: [4,20,77,64] 
    q_neg, q_pos = torch.split(q, q.size(0) // 2, dim=0)
    k_neg, k_pos = torch.split(k, k.size(0) // 2, dim=0)
    v_neg, v_pos = torch.split(v, v.size(0) // 2, dim=0)

    # 1. negative_attn

    scores_neg = torch.matmul(q_neg, k_neg.transpose(-2, -1)) * scale
    attn_weights_neg = torch.softmax(scores_neg, dim=-1)
    attn_output_neg = torch.matmul(attn_weights_neg, v_neg)

    # 2. positive_attn (we do sdsa only on positive branch)

    # 2.1 sdca 
    k_pos_reshaped = k_pos.permute(0, 1, 3, 2)  # 变换维度顺序
    k_plus = torch.cat((k_pos_reshaped, k_pos_reshaped), dim=3)  # 在最后一维拼接
    v_plus = torch.cat((v_pos, v_pos), dim=2)  # 在最后一维拼接
                                       
    # k_plus = torch.cat(tuple(k_pos.transpose(-2, -1)), dim=2).unsqueeze(0).repeat(k_pos.size(0),1,1,1) # 𝐾+ = [𝐾1 ⊕ 𝐾2 ⊕ . . . ⊕ 𝐾𝑁 ]
    # v_plus = torch.cat(tuple(v_pos), dim=1).unsqueeze(0).repeat(v_pos.size(0),1,1,1) # 𝑉+ = [𝑉1 ⊕ 𝑉2 ⊕ . . . ⊕ 𝑉𝑁 ]

    # visuialize: qkv_pos_save
    if unet_controller is not None and unet_controller.save_intermediate_result and (unet_controller.current_time_step + 1) % 5 == 0:
        qkv_save_dir = os.path.join(unet_controller.result_save_dir,f"cross_qkv")
        if os.path.exists(qkv_save_dir) is False:
            os.mkdir(qkv_save_dir)

        for prompt_index in range(len(unet_controller.prompts)):

            q_pos_save = torch.sum(q_pos[prompt_index],dim=0)
            q_pos_save = torch.sum(q_pos_save,dim=1)
            q_pos_save = q_pos_save.view(int(q_pos_save.size(0)**0.5),int(q_pos_save.size(0)**0.5))
            save_tensor_as_image(q_pos_save,f'{qkv_save_dir}/q_prompt{prompt_index}_{unet_controller.current_time_step}.jpg')
    

    # 2.2 apply mask
    if unet_controller is not None:
        scores_pos = torch.matmul(q_pos, k_plus) * scale

        # visuialize: attn_map_save
        if unet_controller.save_cross_attn_map and unet_controller.save_intermediate_result and unet_controller.current_time_step == 49:
            map_save_dir = os.path.join(unet_controller.result_save_dir,f"cross_attn_map")
            if os.path.exists(map_save_dir) is False:
                os.mkdir(map_save_dir)
            side_length = int(scores_pos.size(3)/len(unet_controller.prompts))
            attn_map_save = torch.sum(scores_pos, dim=1)
            for prompt_index in range(len(unet_controller.prompts)):
                attn_map_prompt = attn_map_save[prompt_index,:,prompt_index*side_length:(prompt_index+1)*side_length]
                save_tensor_as_image(attn_map_prompt,f'{map_save_dir}/{prompt_index}_{unet_controller.current_time_step}.jpg')
            unet_controller.save_cross_attn_map = False
        

        # 2.2.1 apply dropout mask
        dropout_mask = gen_dropout_mask(scores_pos.shape, unet_controller, unet_controller.sdca_dropout) # eg: [a,1024,154]   

        # 2.2.2 weak story_self_attn_map
        if unet_controller.ref_prompts is not None and unet_controller.weak_cross_attn_story_prompt:
            weak_story_prompt_cross_attn_map(0,dropout_mask,unet_controller)

        # 2.2.3 apply embeds mask (To Do)

        mask = dropout_mask

        # visuialize: attn_mask_save
        if unet_controller.save_cross_attn_mask and unet_controller.save_intermediate_result and unet_controller.current_time_step == 49:
            map_save_dir = os.path.join(unet_controller.result_save_dir,f"cross_attn_mask")
            if os.path.exists(map_save_dir) is False:
                os.mkdir(map_save_dir)
            for prompt_index in range(len(unet_controller.prompts)):
                save_tensor_as_image(mask[prompt_index],f'{map_save_dir}/{prompt_index}_{unet_controller.current_time_step}.jpg')
            unet_controller.save_cross_attn_mask = False # only save once

        mask = mask.unsqueeze(1).repeat(1,scores_pos.size(1),1,1)
        attn_weights_pos = torch.softmax(scores_pos + torch.log(mask), dim=-1)

    else:
        scores_pos = torch.matmul(q_pos, k_plus) * scale
        attn_weights_pos = torch.softmax(scores_pos, dim=-1)


    attn_output_pos = torch.matmul(attn_weights_pos, v_plus)
    # 3. combine
    attn_output = torch.cat((attn_output_neg, attn_output_pos), dim=0)

    return attn_output


def get_token_indices_by_keyword(keyword, prompt, tokenizer, get_eot=False):
    words = keyword.split()
    token_prompt = prompt2tokens(tokenizer, prompt)

    token_words = [word+'</w>' for word in words]
    if get_eot:
        token_words.append('<|endoftext|>')
    token_indices = []
    for index, token in enumerate(token_prompt):
        if token in token_words:
            token_indices.append(index)
    
    return token_indices


def weak_story_prompt_cross_attn_map(story_index,dropout_mask,unet_controller: Optional[UNetController]):
    story_index = 0
    prompt_num = len(unet_controller.prompts)
    single_prompt_length = int(dropout_mask.size(2)/prompt_num)
    weak_mask = torch.bernoulli(torch.full((dropout_mask.size(1),single_prompt_length), 1 -  unet_controller.cross_attn_story_prompt_dropout, dtype=unet_controller.torch_dtype, device=unet_controller.device))        
    dropout_mask[story_index][:,single_prompt_length*story_index:single_prompt_length*(story_index+1)] =  weak_mask


def weak_story_prompt_self_attn_map(story_index,dropout_mask,unet_controller: Optional[UNetController]): # eg. dropout_mask [a,1024,2048]
    story_index = 0
    prompt_num = len(unet_controller.prompts)
    single_prompt_length = int(dropout_mask.size(2)/prompt_num)


    if unet_controller.keyword_list is not None and unet_controller.total_avg_attn_map is not None:
        for keyword in unet_controller.keyword_list:
            mask_keyword = get_threshold_cross_attn_map_by_keyword(unet_controller,keyword,story_index,single_prompt_length,unet_controller.device,unet_controller.torch_dtype) # eg. [1024]   
            mask_keyword = torch.tensor(1.0, dtype=unet_controller.torch_dtype, device=unet_controller.device) - mask_keyword  
            mask_keyword_horizontal = mask_keyword.unsqueeze(1).repeat(1,single_prompt_length)
            # mask_keyword_horizontal = mask_keyword.unsqueeze(0).repeat(single_prompt_length,1)
            dropout_mask[story_index][:,single_prompt_length*story_index:single_prompt_length*(story_index+1)] = torch.logical_and(dropout_mask[story_index][:,single_prompt_length*story_index:single_prompt_length*(story_index+1)],
                                                                                                                                   mask_keyword_horizontal).to(unet_controller.torch_dtype)
    
    weak_mask = torch.bernoulli(torch.full((dropout_mask.size(1),single_prompt_length), 1 - unet_controller.self_attn_story_prompt_dropout, dtype=unet_controller.torch_dtype, device=unet_controller.device))        
    dropout_mask[story_index][:,single_prompt_length*story_index:single_prompt_length*(story_index+1)] =  torch.logical_or(dropout_mask[story_index][:,single_prompt_length*story_index:single_prompt_length*(story_index+1)],
                                                                                                                           weak_mask).to(unet_controller.torch_dtype)


def gen_cross_attn_mask_2(unet_controller: Optional[UNetController], mask_side_length):
    prompt_num = len(unet_controller.prompts)
    mask_list = []
    for prompt_index in range(prompt_num):
        mask = torch.ones((mask_side_length,mask_side_length * prompt_num), device=unet_controller.device, dtype=unet_controller.torch_dtype)

        or_mask_keyword_list = None
        for keyword in unet_controller.keyword_list:
  
            mask_keyword = get_threshold_cross_attn_map_by_keyword(unet_controller,keyword,prompt_index,mask_side_length,unet_controller.device,unet_controller.torch_dtype)           
            mask_keyword_horizontal = mask_keyword.unsqueeze(1).repeat(1,mask_side_length) # eg. [1024,1024]

            mask_keyword_list = []
            for remain_prompt_index in range(prompt_num - 1):
                current_prompt_index = (prompt_index + 1 + remain_prompt_index) % prompt_num
                remain_mask_keyword = get_threshold_cross_attn_map_by_keyword(unet_controller,keyword,current_prompt_index,mask_side_length,unet_controller.device,unet_controller.torch_dtype)

                cross_keyword_vertical = remain_mask_keyword.unsqueeze(0).repeat(mask_side_length,1)
                mask_keyword = mask_keyword_horizontal * cross_keyword_vertical
                mask_keyword_list.append(mask_keyword)


            if or_mask_keyword_list is None:
                or_mask_keyword_list = mask_keyword_list
            else:
                for list_index in range(len(or_mask_keyword_list)):
                    or_mask_keyword_list[list_index] = logical_or(or_mask_keyword_list[list_index],mask_keyword_list[list_index])
                    
        for mask_index in range(len(or_mask_keyword_list)):
            assign_index_start = ((prompt_index + 1 + mask_index) * mask_side_length) % (mask_side_length * prompt_num)
            assign_index_end = assign_index_start + mask_side_length
            mask[:,assign_index_start:assign_index_end] *= or_mask_keyword_list[mask_index]
        
        mask_list.append(mask)
    mask = torch.stack(mask_list,dim=0)
    return mask


def gen_cross_attn_mask_1(unet_controller: Optional[UNetController], mask_side_length):
    prompt_num = len(unet_controller.prompts)
    mask_list = []
    for prompt_index in range(prompt_num):
        mask = torch.ones((mask_side_length,mask_side_length * prompt_num), device=unet_controller.device, dtype=unet_controller.torch_dtype)

        or_mask_keyword_list = None
        for keyword in unet_controller.keyword_list:
  
            mask_keyword = get_threshold_cross_attn_map_by_keyword(unet_controller,keyword,prompt_index,mask_side_length,unet_controller.device,unet_controller.torch_dtype)
            mask_keyword_horizontal = mask_keyword.unsqueeze(1).repeat(1,mask_side_length)

            mask_keyword_list = []
            for remain_prompt_index in range(prompt_num - 1):
                current_prompt_index = (prompt_index + 1 + remain_prompt_index) % prompt_num
                
                remain_mask_keyword_all = None
                for keyword_loop2 in unet_controller.keyword_list:
                    remain_mask_keyword = get_threshold_cross_attn_map_by_keyword(unet_controller,keyword_loop2,current_prompt_index,mask_side_length,unet_controller.device,unet_controller.torch_dtype)                   
                    if remain_mask_keyword_all is None:
                        remain_mask_keyword_all = remain_mask_keyword
                    else:
                        remain_mask_keyword_all = logical_or(remain_mask_keyword_all, remain_mask_keyword)

                cross_keyword_vertical = remain_mask_keyword_all.unsqueeze(0).repeat(mask_side_length,1)
                mask_keyword = mask_keyword_horizontal * cross_keyword_vertical
                mask_keyword_list.append(mask_keyword)


            if or_mask_keyword_list is None:
                or_mask_keyword_list = mask_keyword_list
            else:
                for list_index in range(len(or_mask_keyword_list)):
                    or_mask_keyword_list[list_index] = logical_or(or_mask_keyword_list[list_index],mask_keyword_list[list_index])
                    
        for mask_index in range(len(or_mask_keyword_list)):
            assign_index_start = ((prompt_index + 1 + mask_index) * mask_side_length) % (mask_side_length * prompt_num)
            assign_index_end = assign_index_start + mask_side_length
            mask[:,assign_index_start:assign_index_end] *= or_mask_keyword_list[mask_index]
        
        mask_list.append(mask)
    mask = torch.stack(mask_list,dim=0)
    return mask


def gen_cross_attn_mask_3(unet_controller: Optional[UNetController], mask_side_length):
    prompt_num = len(unet_controller.prompts)
    mask_list = []
    sqrt_side_length = int(mask_side_length ** 0.5)
    for prompt_index in range(prompt_num):
        mask = torch.ones((mask_side_length,mask_side_length * prompt_num), device=unet_controller.device, dtype=unet_controller.torch_dtype)
        
        or_mask_keyword_list = None
        for keyword in unet_controller.keyword_list:

            mask_keyword_list = []
            for remain_prompt_index in range(prompt_num - 1):
                current_prompt_index = (prompt_index + 1 + remain_prompt_index) % prompt_num
                current_cross_mask = get_threshold_cross_attn_map_by_keyword(unet_controller,keyword,current_prompt_index,mask_side_length,unet_controller.device,unet_controller.torch_dtype)
                current_cross_mask = interpolate_tensor(current_cross_mask.unsqueeze(0), scale_factor=int(mask_side_length / sqrt_side_length)).squeeze()
                current_cross_mask = current_cross_mask.view(mask_side_length,mask_side_length)

                current_cross_mask = torch.where(current_cross_mask != 0, torch.tensor(1.0, device=unet_controller.device, dtype=unet_controller.torch_dtype), torch.tensor(0.0, device=unet_controller.device, dtype=unet_controller.torch_dtype))
                current_cross_mask = self_attn_map_reshape_back(current_cross_mask)
                mask_keyword_list.append(current_cross_mask)
                # save_tensor_as_image(current_cross_mask,f'{unet_controller.result_save_dir}/test.jpg')

            if or_mask_keyword_list is None:
                or_mask_keyword_list = mask_keyword_list
            else:
                for list_index in range(len(or_mask_keyword_list)):
                    or_mask_keyword_list[list_index] = logical_or(or_mask_keyword_list[list_index],mask_keyword_list[list_index])

        for mask_index in range(len(or_mask_keyword_list)):
            assign_index_start = ((prompt_index + 1 + mask_index) * mask_side_length) % (mask_side_length * prompt_num)
            assign_index_end = assign_index_start + mask_side_length
            mask[:,assign_index_start:assign_index_end] *= or_mask_keyword_list[mask_index]
        
        mask_list.append(mask)
    mask = torch.stack(mask_list,dim=0)
    return mask


def gen_cross_attn_mask_0(unet_controller: Optional[UNetController], mask_side_length):
    keyword_attn_mask_list = []
    
    for prompt_index in range(len(unet_controller.prompts)):
        attn_mask = None
        
        for keyword in unet_controller.keyword_list:

            keyword_attn_map = get_threshold_cross_attn_map_by_keyword(unet_controller,keyword,prompt_index,mask_side_length,unet_controller.device,unet_controller.torch_dtype)

            if attn_mask is None:
                attn_mask = keyword_attn_map
            else:
                attn_mask = logical_or(attn_mask, keyword_attn_map).to(unet_controller.torch_dtype)

        keyword_attn_mask_list.append(attn_mask)

    keyword_attn_map = torch.cat(keyword_attn_mask_list, dim=0)

    final_keyword_attn_mask_list = []
    for prompt_index in range(len(unet_controller.prompts)):
        temp = keyword_attn_map.clone()

        start = prompt_index * mask_side_length
        end = (prompt_index + 1) * mask_side_length
        temp[start:end] = 1
        final_keyword_attn_mask_list.append(temp)

    final_keyword_attn_mask = torch.stack(final_keyword_attn_mask_list,dim=0)

    final_keyword_attn_mask = final_keyword_attn_mask.unsqueeze(1).repeat(1,mask_side_length,1)

    return final_keyword_attn_mask


def get_threshold_cross_attn_map_by_keyword(unet_controller: Optional[UNetController], keyword, prompt_index, mask_length, device, dtype):
    keyword_attn_map = unet_controller.get_cross_attn_map_by_keyword(keyword,prompt_index,get_positive_branch=True,reshape=False,sum=True).cpu().float() # [a, 1024]
    if mask_length != keyword_attn_map.size(1):
        keyword_attn_map = interpolate_tensor(keyword_attn_map,scale_factor=(mask_length/keyword_attn_map.size(1))**0.5)

    if keyword_attn_map.size(0) == 0: # when keyword not found, set mask to all zero
        keyword_attn_map = torch.zeros(mask_length)
    else:
        if unet_controller.threshold_mode == 'ostu':
            keyword_attn_map = otsu_threshold(keyword_attn_map) # [a, 1024]
        elif unet_controller.threshold_mode == 'manual':
            keyword_attn_map = manual_threshold(keyword_attn_map, unet_controller.cross_attn_threshold) # [a, 1024]
        else:
            exit("threshold_mode error")
        keyword_attn_map = logical_and(keyword_attn_map) # [1024]
    keyword_attn_map = keyword_attn_map.to(device)
    keyword_attn_map = keyword_attn_map.to(dtype)
    return keyword_attn_map

# 𝑀+_𝑖 = [𝑀1 . . . 𝑀𝑖−1 ⊕ 1 ⊕ 𝑀𝑖+1 . . . 𝑀𝑁]
def gen_dropout_mask(out_shape, unet_controller: Optional[UNetController], drop_out):
    gen_length = out_shape[3]
    attn_map_side_length = out_shape[2]

    prompt_num = len(unet_controller.prompts)
    ref_prompt_num = len(unet_controller.ref_prompts) if unet_controller.ref_prompts is not None else 0
    assert(ref_prompt_num <= 1)
    mask_list = []
    
    for prompt_index in range(prompt_num):
        start = prompt_index * int(gen_length / prompt_num)
        end = (prompt_index + 1) * int(gen_length / prompt_num)
    
        if ref_prompt_num == 1 and prompt_index != 0: # if use ref_prompt, only this first prompt need to use sdsa, set others' mask to all zero
            mask = torch.zeros((attn_map_side_length,gen_length), device=unet_controller.device, dtype=unet_controller.torch_dtype)
        else:
            mask = torch.bernoulli(torch.full((attn_map_side_length,gen_length), 1 - drop_out, dtype=unet_controller.torch_dtype, device=unet_controller.device))        
        mask[:, start:end] = 1

        mask_list.append(mask)

    concatenated_mask = torch.stack(mask_list, dim=0)
    return concatenated_mask


def load_pipe_from_path(model_path, device, torch_dtype, variant):
    model_name = model_path.split('/')[-1]
    if model_path.split('/')[-1] == 'playground-v2.5-1024px-aesthetic':
        scheduler = EDMDPMSolverMultistepScheduler.from_pretrained(model_path, subfolder="scheduler", torch_dtype=torch_dtype, variant=variant,)
    else:
        scheduler = EulerDiscreteScheduler.from_pretrained(model_path, subfolder="scheduler", torch_dtype=torch_dtype, variant=variant,)
    
    if model_path.split('/')[-1] == 'Juggernaut-X-v10':
        variant = None

    vae = AutoencoderKL.from_pretrained(model_path, subfolder="vae", torch_dtype=torch_dtype, variant=variant,)
    tokenizer = CLIPTokenizer.from_pretrained(model_path, subfolder="tokenizer", torch_dtype=torch_dtype, variant=variant,)
    tokenizer_2 = CLIPTokenizer.from_pretrained(model_path, subfolder="tokenizer_2", torch_dtype=torch_dtype, variant=variant,)
    text_encoder = CLIPTextModel.from_pretrained(model_path, subfolder="text_encoder", torch_dtype=torch_dtype, variant=variant,)
    text_encoder_2 = CLIPTextModelWithProjection.from_pretrained(model_path, subfolder="text_encoder_2", torch_dtype=torch_dtype, variant=variant,)
    unet_new = UNet2DConditionModel.from_pretrained(model_path, subfolder="unet", torch_dtype=torch_dtype, variant=variant,)
    
    pipe = pipeline_stable_diffusion_xl.StableDiffusionXLPipeline(
        vae=vae,
        text_encoder=text_encoder,
        text_encoder_2=text_encoder_2,
        tokenizer=tokenizer,
        tokenizer_2=tokenizer_2,
        unet=unet_new,
        scheduler=scheduler,
    )
    pipe.to(device)

    return pipe, model_name

def normalize_image(image):
    mean = image.mean()
    std = image.std()
    return (image - mean) / std


def save_tensor_as_image(tensor, save_path, save_size=None, normalize=True, overwrite=True):
    tensor = tensor.cpu().float()
    if tensor.dim() == 3:
        tensor = tensor[0]

    if normalize:
        # Normalize the tensor to the range [0, 1]
        normalized_tensor = (tensor - torch.min(tensor)) / (torch.max(tensor) - torch.min(tensor))

        # Convert the tensor to a NumPy array and scale to 0-255
        numpy_array = (normalized_tensor * 255).to(torch.uint8).numpy()
    else:
        numpy_array = tensor.to(torch.uint8).numpy()
    # Convert the NumPy array to a PIL Image
    image = Image.fromarray(numpy_array)
    if save_size is not None:
        image = image.resize(save_size, Image.BILINEAR)
    if os.path.exists(save_path) is False or overwrite is True:
        image.save(save_path)


def iter_gen_story(id_prompt, movement_prompt_list, pipe, unet_controller: Optional[UNetController], combine_num=2):
    story_images = []
    if unet_controller is not None:
        seed = unet_controller.get_variable('seed') if unet_controller.get_variable('seed') is not None else random.randint(1,10000)
    
    generate = torch.Generator(device=unet_controller.device).manual_seed(seed)
    for i in range(len(movement_prompt_list) - combine_num + 1):
        gen_prompts = []
        for j in range(combine_num):
            current_prompt = f'{id_prompt} {movement_prompt_list[i+j]}'
            gen_prompts.append(current_prompt)

        images = pipe(gen_prompts, generator=generate, unet_controller=unet_controller).images
        
        if len(story_images) == 0:
            for k in range(combine_num - 1):
                story_images.append(images[k])
        story_images.append(images[combine_num-1])
        unet_controller.reinit_for_next_prompts()
        torch.cuda.empty_cache()


    image_array_list = [np.array(pil_img) for pil_img in story_images]
    # Concatenate images horizontally
    story_image = np.concatenate(image_array_list, axis=1)
    story_image = Image.fromarray(story_image.astype(np.uint8))

    return story_images, story_image

def movement_gen_story_exp(id_prompt, movement_prompt_list, pipe, unet_controller: Optional[UNetController]):
    story_images = []
    unet_controller.set_ref_prompts([id_prompt])


    for index, movement in enumerate(movement_prompt_list):
        gen_propmts = [id_prompt + movement]

        unet_controller.set_variable('movements_to_weaken',movement_prompt_list[:index] + movement_prompt_list[index+1:])
        unet_controller.set_variable('movement_to_strengthen',[movement])
        generate = torch.Generator(device=unet_controller.device).manual_seed(unet_controller.get_variable('seed')) # keep ramd noise to be the same
        
        images = pipe(gen_propmts, height=unet_controller.height, width=unet_controller.width, generator=generate, unet_controller=unet_controller).images
        # if index == 0:
        #     story_images.append(images[1])
        story_images.append(images[0])
        unet_controller.reinit_for_next_prompts()
        # torch.cuda.empty_cache()


    image_array_list = [np.array(pil_img) for pil_img in story_images]
    # Concatenate images horizontally
    story_image = np.concatenate(image_array_list, axis=1)
    story_image = Image.fromarray(story_image.astype(np.uint8))

    return story_images, story_image

def movement_gen_story_slide_windows(id_prompt, movement_prompt_list, pipe, window_length, seed, unet_controller: Optional[UNetController]):
    if window_length > len(movement_prompt_list):
        window_length = len(movement_prompt_list)
    if window_length <= 1:
        window_length = 2
    
    movement_lists = circular_sliding_windows(movement_prompt_list, window_length)
    story_images = []

    if unet_controller is not None:
        unet_controller.set_ref_prompts([id_prompt])

    generate = torch.Generator().manual_seed(seed)

    for index, movement in enumerate(movement_prompt_list):
        if unet_controller is not None:
            unet_controller.set_variable('movements_to_weaken',movement_lists[index][1:])
            unet_controller.set_variable('movement_to_strengthen',movement_lists[index][0])
            unet_controller.set_variable('id_prompt',id_prompt)
            if unet_controller.get_variable("single") == True:
                gen_propmts = f'{id_prompt} {" ".join(movement_lists[index])}'
            else:
                gen_propmts = [f'{id_prompt} {" ".join(movement_lists[index])}', id_prompt]

            if unet_controller.get_variable("original") == True:
                print("original")
                gen_propmts = f'{id_prompt} {movement}'

        else:
            gen_propmts = f'{id_prompt} {movement}'


        if unet_controller is not None and unet_controller.get_variable("use_same_init_noise") is True:     
            generate = torch.Generator().manual_seed(seed)

        images = pipe(gen_propmts, generator=generate, unet_controller=unet_controller).images
        story_images.append(images[0])

        if unet_controller is not None:
            unet_controller.reinit_for_next_prompts()

    image_array_list = [np.array(pil_img) for pil_img in story_images]
    # Concatenate images horizontally
    story_image = np.concatenate(image_array_list, axis=1)
    story_image = Image.fromarray(story_image.astype(np.uint8))

    return story_images, story_image

def movement_gen_story_slide(id_prompt, movement_prompt_list, pipe, unet_controller: Optional[UNetController]):
    story_images = []
    unet_controller.set_ref_prompts([id_prompt])


    for index, movement in enumerate(movement_prompt_list):
        movement_to_weak = movement_prompt_list[(index+1) % len(movement_prompt_list)]
        story_prompt = f'{id_prompt} {movement} {movement_to_weak}'
        gen_propmts = [story_prompt,id_prompt]
        unet_controller.set_variable('movements_to_weaken',[movement_to_weak])
        unet_controller.set_variable('movement_to_strengthen',[movement])
        generate = torch.Generator(device=unet_controller.device).manual_seed(unet_controller.get_variable('seed')) # keep ramd noise to be the same
        
        images = pipe(gen_propmts, height=unet_controller.height, width=unet_controller.width, generator=generate, unet_controller=unet_controller).images
        story_images.append(images[0])
        unet_controller.reinit_for_next_prompts()
        # torch.cuda.empty_cache()


    image_array_list = [np.array(pil_img) for pil_img in story_images]
    # Concatenate images horizontally
    story_image = np.concatenate(image_array_list, axis=1)
    story_image = Image.fromarray(story_image.astype(np.uint8))

    return story_images, story_image


def get_text_embedding_combine(id_prompt, movement_prompt_list, pipe, unet_controller: Optional[UNetController]):
    unet_controller.prompt_embeds_mode = "get_tb_comb"
    story_prompt = id_prompt + ' ' + ' '.join(movement_prompt_list)
    unet_controller.set_variable('id',id_prompt)
    unet_controller.set_variable("movement_list",movement_prompt_list)
    pipe(story_prompt, height=unet_controller.height, width=unet_controller.width, unet_controller=unet_controller)
    return unet_controller.get_variable("embedding_list")


def get_text_embedding_sep(id_prompt, movement_prompt_list, pipe, unet_controller: Optional[UNetController]):
    unet_controller.prompt_embeds_mode = "get_tb_single"
    embedding_list = []
    unet_controller.set_variable('id',id_prompt)

    for index, movement in enumerate(movement_prompt_list):
        story_prompt = f'{id_prompt} {movement}'
        unet_controller.set_variable('movement',movement)
        generate = torch.Generator(device=unet_controller.device).manual_seed(unet_controller.get_variable('seed')) # keep ramd noise to be the same
        pipe(story_prompt, height=unet_controller.height, width=unet_controller.width, generator=generate, unet_controller=unet_controller)
        embedding_list.append(unet_controller.get_variable('embedding'))

    return embedding_list


def baseline_gen_story(id_prompt, movement_prompt_list, pipe, unet_controller: Optional[UNetController]):
    story_images = []

    generate = torch.Generator(device=unet_controller.device).manual_seed(unet_controller.get_variable('seed')) # keep ramd noise to be the same

    for index, movement in enumerate(movement_prompt_list):
        gen_propmts = f"{id_prompt} {movement}"

        images = pipe(gen_propmts, height=unet_controller.height, width=unet_controller.width, generator=generate).images
        story_images.append(images[0])
        # torch.cuda.empty_cache()


    image_array_list = [np.array(pil_img) for pil_img in story_images]
    # Concatenate images horizontally
    story_image = np.concatenate(image_array_list, axis=1)
    story_image = Image.fromarray(story_image.astype(np.uint8))

    return story_images, story_image


def otsu_threshold(input_tensor):
    numpy_array = input_tensor.numpy()

    # Normalize to the range [0, 255] for 8-bit representation
    if (numpy_array.max() - numpy_array.min()) != 0 :
        numpy_array_norm = (numpy_array - numpy_array.min()) / (numpy_array.max() - numpy_array.min()) * 255.0
    else:
        numpy_array_norm = numpy_array

    # Convert to uint8
    numpy_array_uint8 = np.uint8(numpy_array_norm)

    # Initialize an empty tensor to store the thresholded results
    thresholded_tensor = torch.zeros_like(input_tensor)

    # Loop over each image in the batch
    for i in range(numpy_array_uint8.shape[0]):
        # Apply Otsu's thresholding using OpenCV to each image
        _, thresholded = cv2.threshold(numpy_array_uint8[i], 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
        
        # Convert thresholded numpy array back to torch tensor
        thresholded = np.squeeze(thresholded)
        thresholded_tensor[i] = torch.tensor(thresholded, dtype=torch.float32)

    # Convert all 255s to 1
    thresholded_tensor[thresholded_tensor == 255] = 1

    return thresholded_tensor


def manual_threshold(input_tensor, threshold):
    output_tensor_list = []
    for tensor_index in range(input_tensor.size(0)):
        
        normalize_tensor = normalize_minmax(normalize(input_tensor[tensor_index]))

        output_tensor = torch.where(normalize_tensor < threshold, torch.tensor(0, dtype=input_tensor[tensor_index].dtype), torch.tensor(1, dtype=input_tensor[tensor_index].dtype))
        length = int(len(output_tensor)**0.5)
        output_tensor = tensor_opening_2d(output_tensor.view(length,length),kernel_size=2).view(-1)
        output_tensor_list.append(output_tensor.unsqueeze(0))
    output_tensor = torch.cat(output_tensor_list,dim=0)

    return output_tensor


def tensor_opening_2d(tensor_image, kernel_size=3, iterations=1):
    """
    Apply morphological opening (erosion followed by dilation) on a 2D PyTorch tensor.
    
    Args:
    tensor_image (torch.Tensor): Input tensor with shape (H, W).
    kernel_size (int): Size of the structuring element used for erosion and dilation.
    iterations (int): Number of iterations for erosion and dilation.

    Returns:
    torch.Tensor: Tensor after applying the opening operation.
    """
    # Ensure the input is a 2D tensor
    if tensor_image.ndimension() != 2:
        raise ValueError("Input tensor must be 2D (H, W)")

    # Convert the tensor to a NumPy array and scale to [0, 255] range
    image_np = tensor_image.numpy() * 255
    image_np = image_np.astype(np.uint8)

    # Create a structuring element (kernel) for morphological operations
    kernel = np.ones((kernel_size, kernel_size), np.uint8)

    # Apply morphological opening (erosion followed by dilation)
    opened_image_np = cv2.morphologyEx(image_np, cv2.MORPH_OPEN, kernel, iterations=iterations)

    # Convert the result back to a PyTorch tensor and scale to [0, 1] range
    result_tensor = torch.from_numpy(opened_image_np / 255.0).float()

    return result_tensor


# do logical_and through first dim
def logical_and(tensor):
    if tensor.size(0) == 1:
        return tensor.squeeze()
    
    result = torch.logical_and(tensor[0], tensor[1])
    for i in range(2, tensor.size(0)):
        result = torch.logical_and(result, tensor[i])
    
    result = result.float()
    return result


# do logical_or through first dim
def logical_or(tensor):
    if tensor.size(0) == 1:
        return tensor
    
    result = torch.logical_or(tensor[0], tensor[1])
    for i in range(2, tensor.size(0)):
        result = torch.logical_or(result, tensor[i])
    
    result = result.float()
    return result


# do logical_or through two tensor
def logical_or(tensor1, tensor2):
    result = torch.logical_or(tensor1, tensor2)    
    result = result.float()
    return result


def interpolate_tensor(tensor, scale_factor=2):
    first_dim = tensor.size(0)
    side_length = (int)(tensor.size(1) ** 0.5)
    tensor = tensor.view(first_dim,side_length,side_length)

    reshaped_tensor = torch.nn.functional.interpolate(tensor.unsqueeze(1), scale_factor=scale_factor, mode='bilinear', align_corners=False)
    reshaped_tensor = reshaped_tensor.view(first_dim,int((side_length*scale_factor)**2))
    return reshaped_tensor


def prompt2tokens(tokenizer, prompt):
    text_inputs = tokenizer(
        prompt,
        padding="max_length",
        max_length=tokenizer.model_max_length,
        truncation=True,
        return_tensors="pt",
    )
    text_input_ids = text_inputs.input_ids
    tokens = []
    for text_input_id in text_input_ids[0]:
        token = tokenizer.decoder[text_input_id.item()]
        tokens.append(token)
    return tokens


def svd(tensor, top_k_rate=0.2):
    if tensor.dim() != 2:
        exit("input tensor shape error")
    
    side_length = tensor.size(0)
    k = int(side_length * top_k_rate)
    U, sigma, Vt = torch.linalg.svd(tensor)
    reconstructed_tensor = torch.matmul(U[:, :k] * sigma[:k], Vt[:k, :])

    return reconstructed_tensor

def punish_wight(tensor, latent_size, alpha=1.0, beta=1.2, calc_similarity=False):
    u, s, vh = torch.linalg.svd(tensor)
    u = u[:,:latent_size]
    zero_idx = int(latent_size * alpha)

    if calc_similarity:
        _s = s.clone()
        _s[zero_idx:] = 0
        _tensor = u @ torch.diag(_s) @ vh
        dist = cdist(tensor[:,0].unsqueeze(0).cpu(), _tensor[:,0].unsqueeze(0).cpu(), metric='cosine')
        print(f'The distance between the word embedding before and after the punishment: {dist}')

    if alpha == -.001:
        s *= (torch.exp(-.001 * s) * beta)  # strengthen objects 
    else:
        s *= torch.exp(-alpha*s)  # suppression EOT 
    tensor = u @ torch.diag(s) @ vh
    return tensor

def swr_single_prompt_embeds(swr_words,prompt_embeds,prompt,tokenizer,alpha=1.0, beta=1.2):
    punish_indices = []

    prompt_tokens = prompt2tokens(tokenizer,prompt)
    
    words_tokens = prompt2tokens(tokenizer,swr_words)
    words_tokens = [word for word in words_tokens if word != '<|endoftext|>' and word != '<|startoftext|>']
    index_of_words = find_sublist_index(prompt_tokens,words_tokens)

    punish_indices.extend([num for num in range(index_of_words, index_of_words+len(words_tokens))])
    punish_indices.extend([index for index, word in enumerate(prompt_tokens) if word == '<|endoftext|>'])
    punish_indices = list(set(punish_indices))
    
    wo_batch = prompt_embeds[punish_indices]
    wo_batch = punish_wight(wo_batch.T.to(float), wo_batch.size(0), alpha=alpha, beta=beta).T.to(prompt_embeds.dtype)

    prompt_embeds[punish_indices] = wo_batch

def swr_single_prompt_embeds_all_in_once(swr_words_list,prompt_embeds,prompt,tokenizer,alpha=1.0, beta=1.2):
    punish_indices = []

    prompt_tokens = prompt2tokens(tokenizer,prompt)
    
    for swr_words in swr_words_list:
        words_tokens = prompt2tokens(tokenizer,swr_words)
        words_tokens = [word for word in words_tokens if word != '<|endoftext|>' and word != '<|startoftext|>']
        index_of_words = find_sublist_index(prompt_tokens,words_tokens)

        punish_indices.extend([num for num in range(index_of_words, index_of_words+len(words_tokens))])
    punish_indices.extend([index for index, word in enumerate(prompt_tokens) if word == '<|endoftext|>'])
    punish_indices = list(set(punish_indices))
    wo_batch = prompt_embeds[punish_indices]
    wo_batch = punish_wight(wo_batch.T.to(float), wo_batch.size(0), alpha=alpha, beta=beta).T.to(prompt_embeds.dtype)

    prompt_embeds[punish_indices] = wo_batch

def weight_single_prompt_embeds(zero_words,prompt_embeds,prompt,tokenizer, weight=1.0,):
    punish_indices = []

    prompt_tokens = prompt2tokens(tokenizer,prompt)
    
    words_tokens = prompt2tokens(tokenizer,zero_words)
    words_tokens = [word for word in words_tokens if word != '<|endoftext|>' and word != '<|startoftext|>']
    index_of_words = find_sublist_index(prompt_tokens,words_tokens)

    punish_indices.extend([num for num in range(index_of_words, index_of_words+len(words_tokens))])
    punish_indices = list(set(punish_indices))
    
    wo_batch = prompt_embeds[punish_indices]
    wo_batch = wo_batch * weight

    prompt_embeds[punish_indices] = wo_batch

def exp(id,movement,prompt_embeds,prompt,tokenizer):
    prompt_tokens = prompt2tokens(tokenizer,prompt)
    movement_tokens = prompt2tokens(tokenizer,movement)
    movement_tokens = [word for word in movement_tokens if word != '<|endoftext|>' and word != '<|startoftext|>']
    index_of_movement = find_sublist_index(prompt_tokens,movement_tokens)

    id_tokens = prompt2tokens(tokenizer,id)
    id_tokens = [word for word in id_tokens] 
    index_of_id = find_sublist_index(prompt_tokens,id_tokens)


    indices = ([num for num in range(index_of_id, index_of_id+len(id_tokens))] + 
               [num for num in range(index_of_movement, index_of_movement+len(movement_tokens))])
    
    return prompt_embeds[indices].unsqueeze(0)

def find_sublist_index(list1, list2):
    for i in range(len(list1) - len(list2) + 1):
        if list1[i:i + len(list2)] == list2:
            return i
    return -1  # If sublist is not found


def normalize(tensor):
    mean = tensor.mean()
    std = tensor.std()
    normalized_tensor = (tensor - mean) / std

    return normalized_tensor

def normalize_minmax(tensor):
    min_val = tensor.min()
    max_val = tensor.max()

    # Normalize the tensor to [0, 1]
    normalized_tensor = (tensor - min_val) / (max_val - min_val)

    return normalized_tensor

def self_attn_map_reshape(attn_map):  # [1024,1024] 
    side_length = int(attn_map.size(0)**0.5)
    attn_map_reshape_list = []
    for i in range(side_length):
        attn_map_reshape = attn_map[i*side_length:(i+1)*side_length,:].view(side_length,side_length,side_length)
        attn_map_reshape = attn_map_reshape.permute(1,0,2)
        attn_map_reshape = attn_map_reshape.reshape(side_length,-1)
        attn_map_reshape_list.append(attn_map_reshape)
    attn_map_reshape = torch.cat(attn_map_reshape_list,dim=0)
    return attn_map_reshape      


def self_attn_map_reshape_back(attn_map_reshape):  # [1024,1024] 
    side_length = int(attn_map_reshape.size(0)**0.5)
    attn_map_list = []
    for i in range(side_length):
        for j in range(side_length):
            attn_map = attn_map_reshape[i*side_length:(i+1)*side_length, j*side_length:(j+1)*side_length]
            attn_map_list.append(attn_map.reshape(side_length*side_length))
    attn_map = torch.stack(attn_map_list)
    return attn_map 


def fourier_filter(x_in: "torch.Tensor", threshold: int, scale: int) -> "torch.Tensor":
    """Fourier filter as introduced in FreeU (https://arxiv.org/abs/2309.11497).

    This version of the method comes from here:
    https://github.com/huggingface/diffusers/pull/5164#issuecomment-1732638706
    """
    x = x_in
    B, C, H, W = x.shape

    # Non-power of 2 images must be float32
    if (W & (W - 1)) != 0 or (H & (H - 1)) != 0:
        x = x.to(dtype=torch.float32)

    # FFT
    x_freq = fftn(x, dim=(-2, -1))
    x_freq = fftshift(x_freq, dim=(-2, -1))

    B, C, H, W = x_freq.shape
    mask = torch.ones((B, C, H, W), device=x.device)

    crow, ccol = H // 2, W // 2
    mask[..., crow - threshold : crow + threshold, ccol - threshold : ccol + threshold] = scale
    x_freq = x_freq * mask

    # IFFT
    x_freq = ifftshift(x_freq, dim=(-2, -1))
    x_filtered = ifftn(x_freq, dim=(-2, -1)).real

    return x_filtered.to(dtype=x_in.dtype)


def apply_freeu(
    resolution_idx: int, hidden_states: "torch.Tensor", res_hidden_states: "torch.Tensor", **freeu_kwargs
) -> Tuple["torch.Tensor", "torch.Tensor"]:
    """Applies the FreeU mechanism as introduced in https:
    //arxiv.org/abs/2309.11497. Adapted from the official code repository: https://github.com/ChenyangSi/FreeU.

    Args:
        resolution_idx (`int`): Integer denoting the UNet block where FreeU is being applied.
        hidden_states (`torch.Tensor`): Inputs to the underlying block.
        res_hidden_states (`torch.Tensor`): Features from the skip block corresponding to the underlying block.
        s1 (`float`): Scaling factor for stage 1 to attenuate the contributions of the skip features.
        s2 (`float`): Scaling factor for stage 2 to attenuate the contributions of the skip features.
        b1 (`float`): Scaling factor for stage 1 to amplify the contributions of backbone features.
        b2 (`float`): Scaling factor for stage 2 to amplify the contributions of backbone features.
    """
    if resolution_idx == 0:
        num_half_channels = hidden_states.shape[1] // 2
        hidden_states[:, :num_half_channels] = hidden_states[:, :num_half_channels] * freeu_kwargs["b1"]
        res_hidden_states = fourier_filter(res_hidden_states, threshold=1, scale=freeu_kwargs["s1"])
    if resolution_idx == 1:
        num_half_channels = hidden_states.shape[1] // 2
        hidden_states[:, :num_half_channels] = hidden_states[:, :num_half_channels] * freeu_kwargs["b2"]
        res_hidden_states = fourier_filter(res_hidden_states, threshold=1, scale=freeu_kwargs["s2"])

    return hidden_states, res_hidden_states


def circular_sliding_windows(lst, w):
    n = len(lst)
    windows = []
    for i in range(n):
        window = [lst[(i + j) % n] for j in range(w)]
        windows.append(window)
    return windows